import os
import numpy as np
import xml.etree.ElementTree as et
import msm.sg.sg2dcs as sg2
import msm.geo3d as g3d
import msm.utilities as utl


class BeamSegment(object):
    """ Class for a beam segment.
    """

    def __init__(self):
        self.kp_in = None  #: Inboard point label
        self.kp_out = None  #: Outboard point label

        self.cs_in = None  #: Inboard cross section
        self.cs_out = None  #: Outboard cross section

        self.twist = 0.  #: Twist
        self.dihedral = 0.  #: Dihedral
        self.sweep = 0.  #: Sweep
        self.local_frame = None  #: Local frame

        self.num_divisions = 1  #: Number of divisions for meshing


class Beam(object):
    """ A slender beam-like structure

    This class is mainly for the GEBT code.
    """

    def __init__(self):
        self.name = ''  #: Name of the beam
        self.fn_gebt_in = ''  #: File name of the GEBT input
        self.fn_gebt_out = ''  #: File name of the GEBT output
        self.fn_gmsh_msh = ''  #: File name of the Gmsh mesh file

        # Analysis
        self.analysis_type = 0  #: Analysis type (GEBT)
        self.max_iteration = 1  #: Max iteration
        self.num_steps = 1  #: Number of analysis steps
        self.num_eigens = 0  #: Number of eigen analysis resutls

        # Design
        self.length = 0.  #: Length of the beam
        self.angular_velocity = []  #: Angular velocity of the rotating beam
        self.linear_velocity = []  #: Linear velocity of the first key point
        self.num_points = 0  #: Number of key points
        self.num_members = 0  #: Number of members
        self.num_materials = 0  #: Number of materials (cross sections)
        self.num_condpoints = 0  #: Number of conditional points
        self.num_condmembers = 0  #: Number of conditional members
        self.num_distrloads = 0  #: Number of distribution loads
        self.num_frames = 0  #: Number of local self.frames
        self.num_initcurvatures = 0  #: Number of initial curvatures
        self.num_timefunctions = 0  #: Number of time functions
        self.stations = []  #: Stations of the beam
        self.members = {}  #: Members
        self.divisions = []  #: Divisions of members
        self.pconditions = []  #: Point conditions (B.C. and loads)
        self.frames = []  #: Local self.frames

        self.points = {}  #: Key points
        self.segments = {}  #: Segments of the beam.

        # Results
        # self.results = None  #: Results of GEBT analysis

    def readPreGEBTIn(self, fn_pregebt_in):
        """ Preprocessor of GEBT

        :param fn_pregebt_in: File name of the PreGEBT main input file (.xml)
        :type fn_pregebt_in: string
        """
        # ----------------------------------------------------------------
        # Parse XML parametric inputs
        tree = et.parse(fn_pregebt_in)
        root = tree.getroot()
        self.name = root.get('name')

        # self.fn_gebt_in = self.name + '_gebt.dat'

        # Read analysis settings
        analysis = root.find('analysis')
        self.analysis_type = int(analysis.find('type').text)
        self.max_iteration = int(analysis.find('max_iteration').text)
        self.num_steps = int(analysis.find('num_steps').text)
        if self.analysis_type == 3:
            self.num_eigens = int(analysis.find('num_eigenvalue').text)

        # Read design
        design = root.find('design')

        self.length = float(design.find('radius').text)

        # if cs_results is None:
        # stations = []
        for st in design.findall('station'):
            station = {}
            st_id = len(self.stations)
            # print st_id

            st_name = st.get('name')
            station['station_name'] = st_name

            # Location
            st_x1 = float(st.find('location').text.strip())
            if (st.find('location').get('unit') == 'percent'):
                st_x1 = st_x1 * self.length
            station['coordinates'] = np.array([st_x1, 0., 0.])

            # Cross section
            cs_name = st.find('cross_section').text
            station['section_name'] = cs_name
            station['template'] = cs_name
            station['cross_section'] = sg2.SGCrossSection(
                st_name + '_' + cs_name)

            # Local frame
            sweep = 0.0
            if st.find('sweep') is not None:
                sweep = float(st.find('sweep').text.strip())
            station['sweep'] = sweep

            dihedral = 0.0
            if st.find('dihedral') is not None:
                dihedral = float(st.find('dihedral').text.strip())
            station['dihedral'] = dihedral

            frame_b = [np.array([1., 0., 0.]), np.array(
                [0., 1., 0.]), np.array([0., 0., 1.])]
            if (sweep != 0.0) or (dihedral != 0.0):
                frame_b = utl.calcFramebFromSweepDihedral(sweep, dihedral)
            station['frame_b'] = frame_b

            # Twist and pitch angle
            twist = 0.0
            if st.find('twist') is not None:
                twist = float(st.find('twist').text.strip())
            station['twist'] = twist

            if st.find('pitch') is not None:
                st_pitch = float(st.find('pitch').text.strip())
            else:
                st_pitch_prev = self.stations[st_id - 1]['pitch']
                l = st_x1 - self.stations[st_id - 1]['coordinates'][0]
                st_pitch = st_pitch_prev + l * twist
            station['pitch'] = st_pitch

            # Discretize
            # if member.find('num_elements') is not None:
            #     ndiv = int(member.find('num_elements').text)
            # else:
            #     mlength = x1s[mid+1] - x1s[mid]
            #     ndiv = int(np.ceil(mlength / ms))
            # cs = {
            #     'mb': mb_name,
            #     'cs': cs_name,
            #     'sw': sweep,
            #     'dh': dihedral,
            #     'b': frame_b,
            #     'ndiv': ndiv
            # }
            self.stations.append(station)
        # return stations
    # else:
        # stations = cs_results
        self.num_points = len(self.stations)

        if self.analysis_type != 0:
            ws = float(design.find('angular_velocity').text.strip())
            ws = [0., 0., ws]
            wtf = [0, 0, 0]
            self.angular_velocity = [ws, wtf]
            # vs = float(design.find('linear_velocity').text.strip())
            # vs = np.array(map(float, vs))
            vs = ws * self.stations[0]['coordinates'][0]
            vs = [0, vs, 0]
            vtf = [0, 0, 0]
            self.linear_velocity = [vs, vtf]

        # Calculate complete point coordinates
        # x1s = [stations[0]['x1'],]
        # x2s = [0.0, ]
        # x3s = [0.0, ]
        # x2p = 0.0
        # x3p = 0.0
        # for i, member in enumerate(members):
        for i in range(1, self.num_points):
            st_prev = self.stations[i - 1]
            st_curr = self.stations[i]
            b1 = st_curr['frame_b'][0]
            # dx1 = x1s[i + 1] - x1s[i]
            dx1 = st_curr['coordinates'][0] - st_prev['coordinates'][0]
            dx2 = b1[1] / (b1[0] / dx1)
            dx3 = b1[2] / (b1[0] / dx1)
            # x2n = x2p + dx2
            # x3n = x3p + dx3
            # x2s.append(x2n)
            # x3s.append(x3n)
            # x2p = x2n
            # x3p = x3n
            st_curr['coordinates'][1] = st_prev['coordinates'][1] + dx2
            st_curr['coordinates'][2] = st_prev['coordinates'][2] + dx3

        # Read point conditions
        # conditions = []
        for condition in design.findall('condition'):
            pid = int(condition.find('point').text)
            if pid < 0:
                pid = self.num_points + pid + 1
            components = list(map(int, condition.find(
                'components').text.strip().split()))
            values = list(map(float, condition.find('values').text.strip().split()))
            self.pconditions.append({
                'point': pid,
                'components': components,
                'values': values
            })
        self.num_condpoints = len(self.pconditions)

        # Calculate C_ab
        # self.frames = []
        frame_a = [
            np.array([1., 0., 0.]),
            np.array([0., 1., 0.]),
            np.array([0., 0., 1.])
        ]
        # for member in members:
        for i in range(1, self.num_points):
            frame_b = self.stations[i]['frame_b']
            Cab = utl.calcCab(frame_a, frame_b)
            self.frames.append(Cab)
        self.num_frames = len(self.frames)

        ms = float(design.find('mesh_size').text)
        # divisions = []
        for i in range(1, self.num_points):
            dx1 = self.stations[i]['coordinates'][0] - \
                self.stations[i - 1]['coordinates'][0]
            self.divisions.append(int(np.ceil(dx1 / ms)))

        # self.num_members = len(members)
        # Assume
        self.num_members = self.num_points - 1
        self.num_materials = self.num_points
        self.num_condmembers = 0
        self.num_distrloads = 0
        self.num_timefunctions = 0
        self.num_initcurvatures = 0

    def writeGEBTIn(self):
        """ Write data to the GEBT input"""
        self.fn_gebt_in = self.name + '_gebt.dat'
        with open(self.fn_gebt_in, 'w') as fout:
            # fout.write('{0:d}  {1:d}  {2:d}\n'.format(self.analysis_type, self.max_iteration, self.num_steps))
            utl.writeFormatIntegers(
                fout, [self.analysis_type, self.max_iteration, self.num_steps], '4d')
            if self.analysis_type != 0:
                utl.writeFormatIntegers(fout, self.angular_velocity[0])
                utl.writeFormatIntegers(fout, self.angular_velocity[1])
                utl.writeFormatIntegers(fout, self.linear_velocity[0])
                utl.writeFormatIntegers(fout, self.linear_velocity[1])
            if self.analysis_type == 3:
                fout.write('{0:8d}\n'.format(self.num_eigens))
            fout.write('\n')

            line = [
                self.num_points, self.num_members, self.num_condpoints,
                self.num_materials, self.num_frames, self.num_condmembers,
                self.num_distrloads, self.num_timefunctions, self.num_initcurvatures
            ]
            utl.writeFormatIntegers(fout, line, '4d')
            fout.write('\n')

            # Write the block of points
            for i in range(self.num_points):
                fout.write('{0:4d}'.format(i + 1))
                # utl.writeFormatFloats(fout, [x1s[i], x2s[i], x3s[i]])
                utl.writeFormatFloats(fout, self.stations[i]['coordinates'])
            fout.write('\n')

            # Write the block of members
            # for i, member in enumerate(members):
            for i in range(1, self.num_points):
                line = [i, i, i + 1, i, i + 1, i, self.divisions[i - 1], 0]
                utl.writeFormatIntegers(fout, line, '4d')
            fout.write('\n')

            # Write the block of point conditions
            for condition in self.pconditions:
                fout.write('{0:4d}\n'.format(condition['point']))
                utl.writeFormatIntegers(fout, condition['components'], '16d')
                utl.writeFormatFloats(fout, condition['values'])
                utl.writeFormatIntegers(
                    fout, np.zeros(6, dtype=np.int8), '16d')
                utl.writeFormatIntegers(
                    fout, np.zeros(6, dtype=np.int8), '16d')
            fout.write('\n')

            # Write the block of cross sectional properties
            # for i, member in enumerate(members):
            for i, st in enumerate(self.stations):
                cs = st['cross_section']
                fout.write('{0:4d}\n'.format(i + 1))
                # tcm = st['tcm']
                # for row in tcm:
                #     utl.writeFormatFloats(fout, row)
                for row in cs.etcm:
                    utl.writeFormatFloats(fout, row)
                fout.write('\n')
                if self.analysis_type != 0:
                    # mass = st['mass']
                    # for row in mass:
                    #     utl.writeFormatFloats(fout, row)
                    for row in cs.mass:
                        utl.writeFormatFloats(fout, row)
                    fout.write('\n')

            # Write the block of self.frames
            for i, frame in enumerate(self.frames):
                fout.write('{0:4d}\n'.format(i + 1))
                for row in frame:
                    utl.writeFormatFloats(fout, row)
                fout.write('\n')

    def writeGmshMsh(self):
        """ Write the Gmsh mesh file for the visualization of the whole blade
        """
        # Write meshing data
        num_nodes_total = len(self.stations)
        num_elements_total = len(self.stations) - 1
        for st in self.stations:
            num_nodes_total += st['cross_section'].num_nodes
            num_elements_total += st['cross_section'].num_elements

        self.fn_gmsh_msh = self.name + '.msh'
        with open(self.fn_gmsh_msh, 'w') as fout:
            # Mesh format
            fout.write('$MeshFormat\n')
            fout.write('2.2  0  8\n')  # version-number  file-type  data-size
            fout.write('$EndMeshFormat\n')
            fout.write('\n')

            # Nodes
            fout.write('$Nodes\n')
            fout.write(str(num_nodes_total) + '\n')
            # Reference line
            for i, st in enumerate(self.stations):
                fout.write('{0:8d}'.format(i + 1))
                utl.writeFormatFloats(fout, np.array(st['coordinates']))
            # Cross sections
            nnodes = len(self.stations)
            for st in self.stations:
                loc = np.array(st['coordinates'])
                cs = st['cross_section']
                for nid, coords in cs.nodes.items():
                    fout.write('{0:8d}'.format(nid + nnodes))
                    utl.writeFormatFloats(fout, np.array(coords) + loc)
                nnodes += cs.num_nodes
            fout.write('$EndNodes\n')
            fout.write('\n')

            # Elements
            fout.write('$Elements\n')
            fout.write(str(num_elements_total) + '\n')
            # Reference line
            for i in range(len(self.stations) - 1):
                line = [i + 1, 1, 2, 1, 1, i + 1, i + 2]
                utl.writeFormatIntegers(fout, line)
            # Cross sections
            nnodes = len(self.stations)
            nelems = len(self.stations) - 1
            for st in self.stations:
                cs = st['cross_section']
                for eid, nodes in cs.elements.items():
                    line = np.array([eid + nelems, 2, 2, 1, 1])
                    line = np.concatenate([line, np.array(nodes) + nnodes])
                    utl.writeFormatIntegers(fout, line)
                nnodes += cs.num_nodes
                nelems += cs.num_elements
            fout.write('$EndElements\n')
            fout.write('\n')

        # Write options
        self.fn_gmsh_msh_opt = self.fn_gmsh_msh + '.opt'
        with open(self.fn_gmsh_msh_opt, 'w') as fout:
            fout.write('General.Axes = ' + str(3) + ';\n')
            fout.write('General.Orthographic = ' + str(0) + ';\n')
            fout.write('Mesh.Lines = ' + str(1) + ';\n')
            fout.write('Mesh.LineWidth = ' + str(1) + ';\n')

    def readGEBTIn(self, fn_gebt_in=''):
        """ Read GEBT input
        """
        if fn_gebt_in != '':
            self.fn_gebt_in = fn_gebt_in
        # results = {}
        lines = []

        with open(self.fn_gebt_in, 'r') as fin:
            for ln, line in enumerate(fin):
                line = line.strip()
                if line == '':
                    continue
                else:
                    lines.append(line.split('#')[0].strip())

        li = 0
        line1 = list(map(int, lines[li].split()))
        self.analysis_type = line1[0]
        self.max_iteration = line1[1]
        self.num_steps = line1[2]
        if self.analysis_type == 0:
            li += 1
        else:
            li += 5
            if self.analysis_type == 3:
                self.num_eigens = int(lines[li])
                li += 1
        line2 = list(map(int, lines[li].split()))
        self.num_points = line2[0]
        self.num_members = line2[1]
        self.num_condpoints = line2[2]
        self.num_materials = line2[3]
        self.num_frames = line2[4]
        self.num_condmembers = line2[5]
        self.num_distrloads = line2[6]
        self.num_timefunctions = line2[7]
        self.num_initcurvatures = line2[8]
        li += 1
        # Read keypoints
        for i in range(li, li + self.num_points):
            temp = lines[i].split()
            self.points[int(temp[0])] = list(map(float, temp[1:]))
        li += self.num_points
        # Read members
        membs = []
        for i in range(li, li + self.num_members):
            temp = lines[i].split()
            self.members[int(temp[0])] = list(map(int, temp[1:]))
        li += self.num_members

        for i in range(self.num_members):
            self.divisions.append(self.members[i+1][-2])

    def readGEBTOutNode(self, lines):
        out = []
        for l in lines:
            out += list(map(float, l.strip().split()))
        return out

    def readGEBTOut(self, fn_gebt_out=''):
        """ Read GEBT results
        """
        if fn_gebt_out == '':
            fn_gebt_out = self.fn_gebt_in + '.out'

        results = []  # for all steps
        points_results = []
        members_results = []

        with open(fn_gebt_out, 'r') as fin:
            all_lines = fin.readlines()

        i = 0
        sid = 0  # at least one step
        mid = 0
        while i < len(all_lines):
            line = all_lines[i].strip()
            if (line == '') or ('===' in line) or ('---' in line):
                i += 1
                continue
            elif ('Step' in line) or ((self.num_steps == 1) and (sid == 0)):
                sid += 1  # step number
                # point results, member results
                results_step = {'points': [], 'members': []}
                mid = 0
                i += 1
                continue
            elif 'Point' in line:
                # Read point result
                out = self.readGEBTOutNode(all_lines[i + 2:i + 5])
                i += 5
                results_step['points'].append(out)
                # points_results.append(out)
                continue
            elif 'Member' in line:
                # Read member result
                mid += 1
                # nelem = members_in[mid - 1][-2]  # number of divisions
                nelem = self.divisions[mid - 1]
                i += 2
                member_out = []
                for j in range(nelem):
                    if self.analysis_type == 0:
                        # Static
                        out = self.readGEBTOutNode(all_lines[i:i + 3])
                        i += 4
                    else:
                        # Dynamic
                        out = self.readGEBTOutNode(all_lines[i:i + 4])
                        i += 5
                    member_out.append(out)
                results_step['members'].append(member_out)
                # members_results.append(member_out)
                if mid == self.num_members:
                    results.append(results_step)
                continue
            i += 1

        return results

    def solvePLECS(self, x1=None, f1=0, f2=0, f3=0, m1=0, m2=0, m3=0):
        """ Solve the static problem of a prismatic, linearly elastic, 
        cantilever beam. Find the three displacements and three rotations 
        of a point x1 given loads f1, f2, f3, m1, m2, m3 applied at the 
        tip.

        Equations (5.59) and (5.61) from the book 
        Nonlinear Composite Beam Theory by D. H. Hodges 
        are used.
        """
        L = self.length
        if x1 is None:
            x1 = L
        F = np.array([[f1, f2, f3, m1, m2, m3]]).T
        e1tilde = np.array([
            [0.0, 0.0, 0.0],
            [0.0, 0.0, -1.0],
            [0.0, 1.0, 0.0]
        ])
        etcm = self.segments[1].cs_in.etcm
        R = etcm[:3, :3]
        Z = etcm[:3, 3:]
        T = etcm[3:, 3:]

        K = np.zeros((6, 6))

        # Equation (5.61)
        K[:3, :3] = (
            x1 * R + 
            (x1 * L - x1**2 / 2.0) * np.dot(Z, e1tilde) -
            x1**2 / 2.0 * np.dot(e1tilde, Z.T) -
            (x1**2 * L / 2.0 - x1**3 / 6.0) * np.dot(e1tilde, np.dot(T, e1tilde))
        )
        K[:3, 3:] = x1 * Z - x1**2 / 2.0 * np.dot(e1tilde, T)

        # Equation (5.59)
        K[3:, :3] = x1 * Z.T + (x1 * L - x1**2 / 2.0) * np.dot(T, e1tilde)
        K[3:, 3:] = x1 * T

        return np.dot(K, F)
