import numpy as np


class Point2D(object):
    """ A point in the plane (x, y)
    """
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.coord = np.array([x, y])

    def __str__(self):
        return '(' + str(self.x) + ', ' + str(self.y) + ')'


class Shape(object):
    """ A shape in the 2D space
    """
    def __init__(self):
        self.area = 0.0
        self.centroid = None
        self.moa1o = None  #: 1st moment of area with respect to axes passing the origin
        self.moa2o = None  #: 2nd moment of area with respect to axes passing the origin
        self.moa2c = None  #: 2nd moment of area with respect to axes passing the centroid


class Rectangular(Shape):
    """ A rectangular
    """
    def __init__(self, bottomleft=None, topright=None, width=0.0, height=0.0, center=(0.0, 0.0)):
        Shape.__init__(self)
        if (bottomleft is not None) and (topright is not None):
            self.bl = Point2D(bottomleft[0], bottomleft[1])
            self.tr = Point2D(topright[0], topright[1])
            self.center = Point2D(
                (self.bl.x + self.tr.x) / 2.0,
                (self.bl.y + self.tr.y) / 2.0
            )
            self.w = self.tr.x - self.bl.x
            self.h = self.tr.y - self.bl.y
        elif (width != 0.0) and (height != 0.0):
            self.center = Point2D(center[0], center[1])
            self.w = width
            self.h = height
            self.bl = Point2D(
                self.center.x - self.w / 2.0,
                self.center.y - self.h / 2.0
            )
            self.tr = Point2D(
                self.center.x + self.w / 2.0,
                self.center.y + self.h / 2.0
            )


class CircularSector(Shape):
    """ A circular sector
    """
    def __init__(self, r, a0, a1, center=(0.0, 0.0)):
        Shape.__init__(self)
        self.r = r  #: Radius
        self.a0 = a0  #: Begining angle in radians
        self.a1 = a1  #: Ending angle in radians
        if isinstance(center, Point2D):
            self.center = center
        else:
            self.center = Point2D(center[0], center[1])  #: Center of the complete circle


class AnnulusSector(Shape):
    """ An annulus sector
    """
    def __init__(self, rin, rout, a0, a1, center=(0.0, 0.0)):
        Shape.__init__(self)
        self.rin = rin
        self.rout = rout
        self.a0 = a0
        self.a1 = a1
        self.center = Point2D(center[0], center[1])
        self.csin = CircularSector(rin, a0, a1, center)
        self.csout = CircularSector(rout, a0, a1, center)


def calcArea(shape):
    """ Calculate the area of the shape
    """
    if isinstance(shape, Rectangular):
        shape.area = shape.w * shape.h
    elif isinstance(shape, CircularSector):
        shape.area = (shape.a1 - shape.a0) / 2.0 * shape.r**2
    elif isinstance(shape, AnnulusSector):
        shape.area = (shape.a1 - shape.a0) / 2.0 * (shape.rout**2 - shape.rin**2)


def calcFirstMomentOfArea(shape):
    """ Calculate the first moment of area of the shape
    """
    if shape.area == 0.0:
        calcArea(shape)
    if isinstance(shape, CircularSector):
        moa1ox = 1. / 3. * shape.r**3 * (np.cos(shape.a0) - np.cos(shape.a1)) + shape.area * shape.center.y
        moa1oy = 1. / 3. * shape.r**3 * (np.sin(shape.a1) - np.sin(shape.a0)) + shape.area * shape.center.x
    elif isinstance(shape, AnnulusSector):
        calcFirstMomentOfArea(shape.csin)
        calcFirstMomentOfArea(shape.csout)
        moa1ox = shape.csout.moa1o[0] - shape.csin.moa1o[0]
        moa1oy = shape.csout.moa1o[1] - shape.csin.moa1o[1]
    shape.moa1o = (moa1ox, moa1oy)


def calcCentroid(shape):
    """ Calculate the centroid of the shape
    """
    if isinstance(shape, Rectangular):
        shape.centroid = shape.center
    else:
        if shape.area == 0.0:
            calcArea(shape)
        if shape.moa1o is None:
            calcFirstMomentOfArea(shape)
        xbar = shape.moa1o[1] / shape.area
        ybar = shape.moa1o[0] / shape.area
        shape.centroid = Point2D(xbar, ybar)


def calcSecondMomentOfArea(shape):
    """ Calculate the second moment of area of the shape
    """
    if shape.area == 0.0:
        calcArea(shape)
    if shape.centroid is None:
        calcCentroid(shape)

    if isinstance(shape, CircularSector):
        moa2cx = 1. / 8. * shape.r**4 * (
            (shape.a1 - shape.a0) - 1. / 2. * (
                np.sin(2. * shape.a1) - np.sin(2. * shape.a0)
            )
        ) - shape.area * (shape.center.y - shape.centroid.y)**2
        moa2cy = 1. / 8. * shape.r**4 * (
            (shape.a1 - shape.a0) + 1. / 2. * (
                np.sin(2. * shape.a1) - np.sin(2. * shape.a0)
            )
        ) - shape.area * (shape.center.x - shape.centroid.x)**2
        moa2ox = moa2cx + shape.area * shape.centroid.y**2
        moa2oy = moa2cy + shape.area * shape.centroid.x**2
    elif isinstance(shape, Rectangular):
        moa2cx = shape.w * shape.h**3 / 12.0
        moa2cy = shape.w**3 * shape.h / 12.0
        moa2ox = moa2cx + shape.area * shape.centroid.y**2
        moa2oy = moa2cy + shape.area * shape.centroid.x**2
    elif isinstance(shape, AnnulusSector):
        calcSecondMomentOfArea(shape.csin)
        calcSecondMomentOfArea(shape.csout)
        moa2ox = shape.csout.moa2o[0] - shape.csin.moa2o[0]
        moa2oy = shape.csout.moa2o[1] - shape.csin.moa2o[1]
        moa2cx = moa2ox - shape.area * shape.centroid.y**2
        moa2cy = moa2oy - shape.area * shape.centroid.x**2

    shape.moa2c = (moa2cx, moa2cy)
    shape.moa2o = (moa2ox, moa2oy)

