import numpy as np
import xml.etree.ElementTree as et
import msm.utilities as utl


class StructureGenome(object):
    """ A structure genome in the theory of MSG
    """

    def __init__(self, name, micro, macro):
        self.name = name  #: Name of the SG
        self.micro = micro  #: Microscopic dimension of the SG (1, 2, 3)
        #: Macroscopic structure (1: beam, 2: plate/shell, 3: solid/block)
        self.macro = macro

        self.fn_gmsh_msh = self.name + '.msh'  #: File name of the Gmsh mesh file

        # Analysis configurations
        #: What analysis
        #: 0 - homogenization (default)
        #: 1 - dehomogenization/localization/recover
        #: 2 - failure (SwiftComp only)
        self.analysis = 0

        #: Physics included in the analysis
        #: 0 - elastic (default)
        #: 1 - thermoelastic
        #: 2 - conduction
        #: 3 - piezoelectric/piezomagnetic
        #: 4 - thermopiezoelectric/thermopiezomagnetic
        #: 5 - piezoelectromagnetic
        #: 6 - thermopiezoelectromagnetic
        self.physics = 0

        #: Macroscopic structural model
        #: 0 - classical (default)
        #: 1 - refined (e.g. generalized Timoshenko)
        self.model = 0

        self.trans_element = True  #: Flag of transformation of elements
        self.uniform_temperature = True  #: Flag of uniform temperature

        # Materials
        self.num_materials = 0  #: Number of materials
        self.materials = {}  #: Materials {mid: Material, ...}

        # Discretization
        self.global_mesh_size = None  #: Global mesh size
        self.degen_element = 0  #: Flag of the type of elements (SC)
        self.num_nodes = 0  #: Number of nodes
        self.num_elements = 0  #: Number of elements
        self.num_slavenodes = 0  #: Number of slave nodes

        self.nodes = {}  #: Nodal coordinates {nid: [y1, y2, y3], ...}
        #: Elemental connectivities {eid: [nid1, nid2, ...], ...}
        self.elements = {}
        self.elementids = []
        self.elements1d = {}
        self.elements2d = {}
        self.elements3d = {}
    
    def writeGmshMsh(self, fn_gmsh_msh=''):
        if fn_gmsh_msh != '':
            self.fn_gmsh_msh = fn_gmsh_msh

        # Find element type id from number of nodes
        elementType = [
            {2: 1, 3: 8, 4: 26, 5: 27},  # 1d elements
            {3: 2, 6: 9, 4: 3, 8: 16, 9: 10},  # 2d elements
            {4: 4, 10: 11, 8: 5, 20: 17},  # 3d elements
        ]
        
        with open(self.fn_gmsh_msh, 'w') as fout:
            # Mesh format
            fout.write('$MeshFormat\n')
            fout.write('2.2  0  8\n')  # version-number  file-type  data-size
            fout.write('$EndMeshFormat\n')
            fout.write('\n')
            
            # Nodes
            fout.write('$Nodes\n')
            fout.write(str(self.num_nodes))
            fout.write('\n\n')
            for nid, coords in self.nodes.items():
                fout.write('{0:8d}'.format(nid))
                utl.writeFormatFloats(fout, coords)
            fout.write('$EndNodes\n')
            fout.write('\n')

            # Elements
            fout.write('$Elements\n')
            fout.write(str(self.num_elements))
            fout.write('\n\n')
            for eid, nodes in self.elements1d.items():
                line = [eid, elementType[0][len(nodes)], 2, 1, 1] + nodes
                utl.writeFormatIntegers(fout, line)
            for eid, nodes in self.elements2d.items():
                line = [eid, elementType[1][len(nodes)], 2, 1, 1] + nodes
                utl.writeFormatIntegers(fout, line)
            for eid, nodes in self.elements3d.items():
                line = [eid, elementType[2][len(nodes)], 2, 1, 1] + nodes
                utl.writeFormatIntegers(fout, line)
            fout.write('$EndElements\n')
            fout.write('\n')

        # Write options
        self.fn_gmsh_msh_opt = self.fn_gmsh_msh + '.opt'
        with open(self.fn_gmsh_msh_opt, 'w') as fout:
            fout.write('General.Axes = ' + str(3) + ';\n')
            fout.write('General.Orthographic = ' + str(0) + ';\n')
            fout.write('Mesh.Lines = ' + str(1) + ';\n')
            fout.write('Mesh.LineWidth = ' + str(1) + ';\n')