import numpy as np
import xml.etree.ElementTree as et
import msm.utilities as utl
from msm.sg.sg import StructureGenome
from msm.material import Material


class SGCrossSection(StructureGenome):
    """ Stores all information of a cross section.

    This object is used to store and pass data of a cross section 
    between VABS/SC, GEBT, and preprocessors.
    """

    def __init__(self, name):
        StructureGenome.__init__(self, name, 2, 1)

        self.name_template = ''  #: Base name of the main input file template
        self.name_baseline = ''  #: Base name of the base lines file
        self.name_layup = ''  #: Base name of the layups file
        self.fn_vabs_in = self.name + '_vabs.dat'  #: File name of the VABS input
        self.fn_sc_in = self.name + '.sc'  #: File name of the SwiftComp input
        # self.fn_gmsh_msh = self.name + '.msh'  #: File name of the Gmsh mesh file

        self.num_layertypes = 0  #: Number of layer types
        self.layertypes = {}  #: Layer types {lid: [mid, ply angle], ...}
        self.element_layertype = {}
        self.element_theta1 = {}

        self.pitch = 0.0  #: Pitch angle of the cross section
        self.k11 = 0.0  #: Initial twist k_{11}
        self.k12 = 0.0  #: Initial curvature k_{12}
        self.k13 = 0.0  #: Initial curvature k_{13}
        self.cos11 = 1.0  #: Obliqueness. Cosine of the angle between y_1 and x_1
        self.cos21 = 0.0  #: Obliqueness. Cosine of the angle between y_2 and x_1

        self.trapeze = 0  #: Flag for trapeze effect
        self.vlasov = 0  #: Flag for Vlasov model

        # Effective properties
        self.mc = None  #: Mass center
        self.gc = None  #: Geometric center
        self.tc = None  #: Tension center
        self.sc = None  #: Shear center

        self.mass = None  #: Mass matrix (6x6)
        self.masc = None  #: Mass matrix at mass center (6x6)
        self.eden = None  #: Effective density (mass per unit length)
        self.ecsm = None  #: Effective classical stiffness matrix (4x4)
        self.eccm = None  #: Effective classical compliance matrix (4x4)
        self.etsm = None  #: Effective timoshenko stiffness matrix (6x6)
        self.etcm = None  #: Effective timoshenko compliance matrix (6x6)

        # Global results for recovery/dehomogenization
        self.gdisplacements = np.zeros(3)  #: Global displacements [u1, u2, u3]
        #: Global rotations [[c11, c12, c13], [c21, c22, c23], [c31, c32, c33]]
        self.grotations = np.eye(3)
        self.gforces = np.zeros(3)  #: Global forces (VABS) [F1, F2, F3]
        self.gmoments = np.zeros(3)  #: Global moments (VABS) [M1, M2, M3]
        #: Global distributed forces (VABS) [[f1, f2, f3], [d1f1, d1f2, d1f3], [d2f1, d2f2, d2f3], [d3f1, d3f2, d3f3]]
        self.gdforces = np.zeros((4, 3))
        #: Global distributed moments (VABS) [[m1, m2, m3], [d1m1, d1m2, d1m3], [d2m1, d2m2, d2m3], [d3m1, d3m2, d3m3]]
        self.gdmoments = np.zeros((4, 3))
        #: Indicate whether the generalized loads are stresses or strains (SwiftComp)
        self.gmeasure = 'stress'
        #: Global straints (SwiftComp) []
        self.gloads = np.zeros(6)

        #: Load for the horizontal axis for failure envelope (SwiftComp)
        self.feaxis1 = ''
        #: Load for the vertical axis for failure envelope (SwiftComp)
        self.feaxis2 = ''
        #: Number of evaluation points along each axis for failure envelope (SwiftComp)
        self.fendiv = 10

    def printResults(self):
        """ Print analysis results
        """
        with np.printoptions(precision=2):
            print(self.etsm)
        # print np.array2string(self.etsm, precision=6, suppress_small=True)

    def updateXMLElementTree(self, analysis):
        """ Update the element tree for PreVABS input

        :param analysis: Analysis step (h, d, f)
        :type analysis: string
        """
        self.root.set('name', self.name)

        if analysis == 'h':
            # Update for homogenization
            # Update includes
            utl.updateXMLElement(
                self.se_include, 'baseline', self.name_baseline)
            utl.updateXMLElement(self.se_include, 'layup', self.name_layup)

            # Update analysis
            utl.updateXMLElement(self.se_analysis, 'model', str(self.model))
            utl.updateXMLElement(
                self.se_analysis, 'oblique_y1', str(self.cos11))
            utl.updateXMLElement(
                self.se_analysis, 'oblique_y2', str(self.cos21))

            # Update general
            utl.updateXMLElement(self.se_general, 'rotate', str(self.pitch))
            utl.updateXMLElement(self.se_general, 'mesh_size',
                                 str(self.global_mesh_size))

            # Update segments

        elif analysis == 'd':
            # Update for recover/dehomogenization
            self.se_recover = self.root.find('recover')
            if self.se_recover is None:
                self.se_recover = et.SubElement(self.root, 'recover')
            utl.updateXMLElement(self.se_recover, 'displacements', '  '.join(
                map(str, self.gdisplacements)))
            utl.updateXMLElement(self.se_recover, 'rotations', '  '.join(
                map(str, self.grotations.flatten())))
            utl.updateXMLElement(self.se_recover, 'forces',
                                 '  '.join(map(str, self.gforces)))
            utl.updateXMLElement(self.se_recover, 'moments',
                                 '  '.join(map(str, self.gmoments)))

            se_distributed = self.se_recover.find('distributed')
            if se_distributed is None:
                se_distributed = et.SubElement(self.se_recover, 'distributed')
            utl.updateXMLElement(se_distributed, 'forces',
                                 '  '.join(map(str, self.gdforces[0])))
            utl.updateXMLElement(se_distributed, 'forces_d1',
                                 '  '.join(map(str, self.gdforces[1])))
            utl.updateXMLElement(se_distributed, 'forces_d2',
                                 '  '.join(map(str, self.gdforces[2])))
            utl.updateXMLElement(se_distributed, 'forces_d3',
                                 '  '.join(map(str, self.gdforces[3])))

            utl.updateXMLElement(se_distributed, 'moments',
                                 '  '.join(map(str, self.gdmoments[0])))
            utl.updateXMLElement(se_distributed, 'moments_d1',
                                 '  '.join(map(str, self.gdmoments[1])))
            utl.updateXMLElement(se_distributed, 'moments_d2',
                                 '  '.join(map(str, self.gdmoments[2])))
            utl.updateXMLElement(se_distributed, 'moments_d3',
                                 '  '.join(map(str, self.gdmoments[3])))

            self.se_dehomo = self.root.find('dehomo')
            if self.se_dehomo is None:
                self.se_dehomo = et.SubElement(self.root, 'dehomo')
            self.se_dehomo.set('measure', self.gmeasure)
            utl.updateXMLElement(self.se_dehomo, 'displacements', '  '.join(
                map(str, self.gdisplacements)))
            utl.updateXMLElement(self.se_dehomo, 'rotations', '  '.join(
                map(str, self.grotations.flatten())))
            utl.updateXMLElement(self.se_dehomo, 'loads',
                                 '  '.join(map(str, self.gloads)))

        elif 'f' in analysis:
            # Update for failure
            self.se_failure = self.root.find('failure')
            if self.se_failure is None:
                self.se_failure = et.SubElement(self.root, 'failure')
            self.se_failure.set('measure', self.gmeasure)
            if analysis == 'fe':
                utl.updateXMLElement(self.se_failure, 'axis1', self.feaxis1)
                utl.updateXMLElement(self.se_failure, 'axis2', self.feaxis2)
                utl.updateXMLElement(
                    self.se_failure, 'divisions', str(self.fendiv))
            elif analysis == 'fi':
                utl.updateXMLElement(
                    self.se_failure, 'loads', ' '.join(map(str, self.gloads)))

    def readPreVABSIn(self, fn_prevabs_in_main, fn_prevabs_in_baseline, fn_prevabs_in_layup):
        """ Read data from a PreVABS input file

        :param fn_prevabs_in_main: File name of the PreVABS main input
        :type fn_prevabs_in_main: string

        :param fn_prevabs_in_baseline: File name of the PreVABS base lines input
        :type fn_prevabs_in_baseline: string

        :param fn_prevabs_in_layup: File name of the PreVABS layups input
        :type fn_prevabs_in_layups: string
        """
        self.fn_in = fn_prevabs_in_main
        #: Element tree object read from PreVABS input
        self.et_main = et.parse(fn_prevabs_in_main)
        #: Root element of the tree
        self.root = self.et_main.getroot()
        #: Element <include> of the cross section
        self.se_include = self.root.find('include')
        #: Element <analysis> of the cross section
        self.se_analysis = self.root.find('analysis')
        if self.se_analysis.find('model') is not None:
            try:
                self.model = int(self.se_analysis.find('model').text.strip())
            except ValueError:
                pass
        if self.se_analysis.find('oblique_y1') is not None:
            try:
                self.cos11 = float(self.se_analysis.find(
                    'oblique_y1').text.strip())
            except ValueError:
                pass
        if self.se_analysis.find('oblique_y2') is not None:
            try:
                self.cos21 = float(self.se_analysis.find(
                    'oblique_y2').text.strip())
            except ValueError:
                pass
        #: Element <general> of the cross section
        self.se_general = self.root.find('general')
        if self.se_general.find('mesh_size') is not None:
            try:
                self.global_mesh_size = float(
                    self.se_general.find('mesh_size').text.strip())
            except ValueError:
                pass
        #: Element <segments> of the cross section
        self.se_segments = self.root.find('segments')

        if self.root.find('recover') is not None:
            self.se_recover = self.root.find('recover')

        return

    def writePreVABSIn(self, fn_prevabs_in):
        """ Write data to a PreVABS input file

        :param fn_prevabs_in: File name of the PreVABS main input
        :type fn_prevabs_in: string
        """
        self.fn_out = fn_prevabs_in
        self.et_main.write(fn_prevabs_in)

        return

    def readVABSIn(self, fn_vabs_in=''):
        """ Read data from the VABS input file

        :param fn_vabs_in: File name of the VABS input file
        :type fn_vabs_in: string
        """
        if fn_vabs_in != '':
            self.fn_vabs_in = fn_vabs_in
        # self.node_coord = {}  # Nodal coordinates      {n1: [x1, x2, x3], ...]}
        # self.elem_connt = {}  # Element connectivities {e1: [n1, n2, n3, n4, ...], ...]}
        # elem_sectn = {}  # Material sections      {'s1': [e1, e2, e3, ...], ...}
        # Material sections      {l1: [e1, e2, e3, ...], ...}
        self.lyrtp_elem = {}
        self.angle_elem = {}  # {a1: [e1, e2, e3, ...], ...}
        self.elem_label = []  # Element labels         [e1, e2, e3, ...]
        self.elem_angle = {}  # {e1: a1, ...}
        self.lyrtp_mate = {}  # {l1: [m1, a1], ...}
        self.mate_propt = {}  # {m1: [type1, p1, p2, ...], ...}

        self.flag_curve = 0
        self.flag_oblique = 0

        count = 0
        count_node = 1
        count_element = 1
        count_layertype = 1
        count_material = 1
        line_material = 0
        head = 3  # at least 3 lines in the head (flag) part
        with open(self.fn_vabs_in, 'r') as fin:
            for li, line in enumerate(fin):
                line = line.strip()
                if line == '\n' or line == '':
                    continue

                count += 1
                line = line.split()
                if count == 1:
                    # format_flag  nlayer
                    line = list(map(int, line))
                    self.flag_format = line[0]
                    self.num_layertypes = line[1]
                    continue
                elif count == 2:
                    # timoshenko_flag  recover_flag  thermal_flag
                    line = list(map(int, line))
                    self.model = line[0]
                    self.flag_analysis = line[1]
                    self.flag_thermal = line[2]
                    continue
                elif count == 3:
                    # curve_flag  oblique_flag  trapeze_flag  vlasov_flag
                    line = list(map(int, line))
                    self.flag_curve = line[0]
                    self.flag_oblique = line[1]
                    self.flag_trapeze = line[2]
                    self.flag_vlasov = line[3]
                    if self.flag_curve == 1:
                        head += 1  # extra line in the head
                    if self.flag_oblique == 1:
                        head += 1  # extra line in the head
                    continue
                elif self.flag_curve != 0 and count <= head:
                    pass
                elif self.flag_oblique != 0 and count <= head:
                    pass
                elif count == (head + 1):
                    # nnode  nelem  nmate
                    line = list(map(int, line))
                    self.num_nodes = line[0]
                    self.num_elements = line[1]
                    self.num_materials = line[2]
                    continue
                elif count_node <= self.num_nodes:
                    # Read nodal coordinates
                    self.nodes[int(line[0])] = [
                        0.0, float(line[1]), float(line[2])]
                    count_node += 1
                    continue
                elif count_element <= self.num_elements:
                    # Read element connectivities
                    eid = int(line[0])
                    self.elem_label.append(eid)
                    temp = [int(i) for i in line[1:] if i != '0']
                    self.elements[eid] = temp
                    self.elements2d[eid] = temp
                    count_element += 1
                    continue
                elif count_element <= (2 * self.num_elements):
                    # Read element theta1
                    if self.flag_format == 1:
                        # new format
                        eid = int(line[0])
                        lyrtp = int(line[1])
                        theta1 = float(line[2])
                    else:
                        # old format
                        eid = int(line[0])
                        mid = int(line[1])
                        theta3 = float(line[2])
                        theta1 = float(line[3])
                        lyrtp = 0
                        for k, v in self.lyrtp_mate.items():
                            if (mid == v[0]) and (theta3 == v[1]):
                                lyrtp = k
                                break
                        if lyrtp == 0:
                            lyrtp = len(self.lyrtp_mate) + 1
                            self.lyrtp_mate[lyrtp] = [mid, theta3]

                    if lyrtp not in self.lyrtp_elem.keys():
                        self.lyrtp_elem[lyrtp] = []
                    self.lyrtp_elem[lyrtp].append(eid)
                    self.element_layertype[eid] = lyrtp
                    # self.elem_angle[eid] = theta1
                    self.element_theta1[eid] = theta1
                    count_element += 1
                    continue
                elif (self.flag_format == 1) and (count_layertype <= self.num_layertypes):
                    # Read layer types
                    ltid = int(line[0])
                    mid = int(line[1])
                    theta3 = float(line[2])
                    self.lyrtp_mate[ltid] = [mid, theta3]
                    self.layertypes[ltid] = [mid, theta3]
                    count_layertype += 1
                    continue
                elif count_material <= self.num_materials:
                    # Read materials
                    if line_material == 0:
                        mid = int(line[0].split(',')[0])
                        mtype = int(line[1])
                        self.materials[mid] = Material(mid)
                        self.materials[mid].type = mtype
                        self.mate_propt[mid] = [mtype]
                        line_material += 1
                        continue

                    if mtype == 0:
                        # isotropic
                        if line_material == 1:
                            e = float(line[0])
                            nu = float(line[1])
                            self.materials[mid].property_elastic += [e, nu]
                            self.mate_propt[mid] += [e, nu]
                            line_material = 99
                            continue
                    elif mtype == 1:
                        # orthotropic
                        if line_material == 1:
                            e = map(float, line)
                            self.materials[mid].property_elastic += e
                            self.mate_propt[mid] += e
                            line_material += 1
                            continue
                        elif line_material == 2:
                            g = map(float, line)
                            self.materials[mid].property_elastic += g
                            self.mate_propt[mid] += g
                            line_material += 1
                            continue
                        elif line_material == 3:
                            nu = map(float, line)
                            self.materials[mid].property_elastic += nu
                            self.mate_propt[mid] += nu
                            line_material = 99
                            continue
                    elif mtype == 2:
                        # anisotropic
                        continue

                    if line_material == 99:
                        dens = float(line[0])
                        self.materials[mid].density = dens
                        self.mate_propt[mid].append(dens)
                        line_material = 0
                        count_material += 1  # next material
                        continue
                    continue
        return

    def writeVABSIn(self, fn_vabs_in=''):
        if fn_vabs_in != '':
            self.fn_vabs_in = fn_vabs_in
        with open(self.fn_vabs_in, 'w') as fout:
            utl.writeFormatIntegers(fout, [1, self.num_layertypes])
            utl.writeFormatIntegers(
                fout, [self.model, self.analysis, self.physics])
            line = [0, 0, self.trapeze, self.vlasov]
            if (self.k11 != 0.0) or (self.k12 != 0.0) or (self.k13 != 0.0):
                line[0] = 1
            if (self.cos11 != 1.0) or (self.cos21 != 0.0):
                line[1] = 1
            utl.writeFormatIntegers(fout, line)
            utl.writeFormatIntegers(
                fout, [self.num_nodes, self.num_elements, self.num_materials])
            fout.write('\n')

            # Nodes
            for nid in range(1, self.num_nodes + 1):
                fout.write('{0:8d}'.format(nid))
                utl.writeFormatFloats(fout, self.nodes[nid][1:])
            fout.write('\n')

            # Elements
            for eid, nodes in self.elements2d.items():
                if (len(nodes) == 3) or (len(nodes) == 6):
                    nodes.insert(3, 0)
                for i in range(9 - len(nodes)):
                    nodes.append(0)
                fout.write('{0:8d}'.format(eid))
                utl.writeFormatIntegers(fout, nodes)
            fout.write('\n')

            # Element, layer type, theta 1
            for eid, theta1 in self.element_theta1.items():
                lid = self.element_layertype[eid]
                fout.write('{0:8d}{1:8d}{2:16.7e}\n'.format(eid, lid, theta1))
            fout.write('\n')

            # Layer types
            for lid, value in self.layertypes.items():
                mid = 0
                mname = value[0]
                for k, v in self.materials.items():
                    if v.name == mname:
                        mid = k
                fout.write('{0:8d}{1:8d}{2:16.7e}\n'.format(
                    lid, mid, value[1]))
            fout.write('\n')

            # Materials
            for mid, mat in self.materials.items():
                mtype = 0
                if (mat.type == 'orthotropic') or (mat.type == 1):
                    mtype = 1
                elif (mat.type == 'anisotropic') or (mat.type == 2):
                    mtype = 2
                utl.writeFormatIntegers(fout, [mid, mtype])
                elastic = mat.property_elastic
                if mtype == 0:
                    utl.writeFormatFloats(fout, elastic)
                elif mtype == 1:
                    utl.writeFormatFloats(fout, elastic[:3])
                    utl.writeFormatFloats(fout, elastic[3:6])
                    utl.writeFormatFloats(fout, elastic[6:])
                elif mtype == 2:
                    pass
                fout.write('{0:16.7e}\n'.format(mat.density))
                fout.write('\n')

    def readVABSOutHomo(self, fn_vabs_in=''):
        """Read VABS homogenization results

        :param fn_vabs_in: VABS input file name (e.g. example.dat)
        :type fn_vabs_in: str
        """
        if fn_vabs_in != '':
            self.fn_vabs_in = fn_vabs_in
        self.fn_vabs_k = self.fn_vabs_in + '.k'
        linesRead = []
        keywordsIndex = {}

        with open(self.fn_vabs_k, 'r') as fin:
            for lineNumber, line in enumerate(fin):
                linesRead.append(line)
                if 'The Mass Center of the Cross Section' in line:
                    keywordsIndex['mc'] = lineNumber
                if 'The Geometric Center of the Cross Section' in line:
                    keywordsIndex['gc'] = lineNumber
                if 'The Neutral Axes (or Tension Center) of the Cross Section' in line:
                    keywordsIndex['tc'] = lineNumber
                if 'The Generalized Shear Center of the Cross Section in the User Coordinate System' in line:
                    keywordsIndex['sc'] = lineNumber
                if 'Mass Per Unit Span' in line:
                    self.eden = float(line.split('=')[-1].strip())
                if 'Classical Stiffness Matrix' in line:
                    keywordsIndex['csm'] = lineNumber
                if 'Classical Flexibility Matrix' in line:
                    keywordsIndex['cfm'] = lineNumber
                if 'Timoshenko Stiffness Matrix' in line:
                    keywordsIndex['tsm'] = lineNumber
                if 'Timoshenko Flexibility Matrix' in line:
                    keywordsIndex['tfm'] = lineNumber
                if 'Mass Matrix' in line:
                    if 'Mass Center' in line:
                        keywordsIndex['massatcenter'] = lineNumber
                    else:
                        keywordsIndex['mass'] = lineNumber

        ln = keywordsIndex['mc']
        self.mc = np.array([
            0.0,
            float(linesRead[ln + 3].split()[-1]),
            float(linesRead[ln + 4].split()[-1])
        ])

        ln = keywordsIndex['gc']
        self.gc = np.array([
            0.0,
            float(linesRead[ln + 3].split()[-1]),
            float(linesRead[ln + 4].split()[-1])
        ])

        ln = keywordsIndex['tc']
        self.tc = np.array([
            0.0,
            float(linesRead[ln + 3].split()[-1]),
            float(linesRead[ln + 4].split()[-1])
        ])

        ln = keywordsIndex['sc']
        self.sc = np.array([
            0.0,
            float(linesRead[ln + 3].split()[-1]),
            float(linesRead[ln + 4].split()[-1])
        ])

        ln = keywordsIndex['mass']
        self.mass = utl.textToMatrix(linesRead[ln + 3:ln + 9])
        try:
            ln = keywordsIndex['massatcenter']
            self.masc = utl.textToMatrix(linesRead[ln + 3:ln + 9])
        except KeyError:
            self.masc = self.mass

        ln = keywordsIndex['csm']
        self.ecsm = utl.textToMatrix(linesRead[ln + 3:ln + 7])
        ln = keywordsIndex['cfm']
        self.eccm = utl.textToMatrix(linesRead[ln + 3:ln + 7])

        try:
            ln = keywordsIndex['tsm']
            self.etsm = utl.textToMatrix(linesRead[ln + 3:ln + 9])
        except KeyError:
            print('No timoshenko stiffness matrix found.')
        try:
            ln = keywordsIndex['tfm']
            self.etcm = utl.textToMatrix(linesRead[ln + 3:ln + 9])
        except KeyError:
            print('No timoshenko flexibility matrix found.')

        return

    def readSCOutHomo(self, fn_sc_in='', macro_dim=1):
        """Read SwiftComp homogenization results.

        :param fn_swiftcomp_in: SwiftComp input file (e.g. example.sc)
        :type fn_swiftcomp_in: string

        :param macro_dim: Dimension of the structural model
        :type macro_dim: int
        """
        if fn_sc_in != '':
            self.fn_sc_in = fn_sc_in
        self.fn_sc_k = self.fn_sc_in + '.k'
        linesRead = []
        keywordsIndex = {}

        with open(self.fn_sc_k, 'r') as fin:
            for lineNumber, line in enumerate(fin):
                linesRead.append(line)
                if 'The Effective Stiffness Matrix' in line:
                    keywordsIndex['The Effective Stiffness Matrix'] = lineNumber
                if 'The Effective Compliance Matrix' in line:
                    keywordsIndex['The Effective Compliance Matrix'] = lineNumber
                if macro_dim == 1:
                    if 'Timoshenko Stiffness' in line:
                        keywordsIndex['Timoshenko Stiffness'] = lineNumber
                    if 'Timoshenko Compliance' in line:
                        keywordsIndex['Timoshenko Compliance'] = lineNumber
                    if 'Shear Center Location' in line:
                        keywordsIndex['Shear Center Location'] = lineNumber
                elif macro_dim == 2:
                    if 'In-Plane Properties' in line:
                        keywordsIndex['In-Plane Properties'] = lineNumber
                    if 'Flexural Properties' in line:
                        keywordsIndex['Flexural Properties'] = lineNumber
                elif macro_dim == 3:
                    if 'The Engineering Constants' in line:
                        keywordsIndex['The Engineering Constants'] = lineNumber
                if 'Effective Density' in line:
                    keywordsIndex['Effective Density'] = lineNumber

        ems = 6  # effective matrix size

        if macro_dim == 1:
            ems = 4
            try:
                indexStiffness = keywordsIndex['Timoshenko Stiffness']
                self.etsm = utl.textToMatrix(
                    linesRead[indexStiffness + 2:indexStiffness + 8])
                # results['ets'] = timoshenkoStiffness
                # for i in range(timoshenkoStiffness.shape[0]):
                #     for j in range(timoshenkoStiffness.shape[1]):
                #         label = 'ets' + str(i + 1) + str(j + 1)
                #         results[label] = timoshenkoStiffness[i][j]
            except KeyError:
                print('No effective timoshenko stiffness matrix found.')
            try:
                indexCompliance = keywordsIndex['Timoshenko Compliance']
                self.etcm = utl.textToMatrix(
                    linesRead[indexCompliance + 2:indexCompliance + 8])
                # results['etc'] = timoshenkoCompliance
                # for i in range(timoshenkoCompliance.shape[0]):
                #     for j in range(timoshenkoCompliance.shape[1]):
                #         label = 'etc' + str(i + 1) + str(j + 1)
                #         results[label] = timoshenkoCompliance[i][j]
            except KeyError:
                print('No effective timoshenko compliance matrix found.')
        # elif macro_dim == 2:
        #     ems = 6
        #     try:
        #         index = keywordsIndex['In-Plane Properties']
        #         for line in linesRead[index + 2:index + 8]:
        #             line = line.strip()
        #             line = line.split('=')
        #             label = line[0].strip().lower()
        #             value = float(line[-1].strip())
        #             results[label] = value
        #     except KeyError:
        #         print 'No in-plane properties found.'
        # elif macro_dim == 3:
        #     ems = 6
        #     try:
        #         indexConstant = keywordsIndex['The Engineering Constants']
        #         for line in linesRead[indexConstant + 2:indexConstant + 11]:
        #             line = line.strip()
        #             line = line.split('=')
        #             label = line[0].strip().lower()
        #             value = float(line[-1].strip())
        #             results[label] = value
        #     except KeyError:
        #         print 'No engineering constants found.'

        try:
            indexStiffness = keywordsIndex['The Effective Stiffness Matrix']
            self.ecsm = utl.textToMatrix(
                linesRead[indexStiffness + 2:indexStiffness + 2 + ems])
            # for i in range(effectiveStiffness.shape[0]):
            #     for j in range(effectiveStiffness.shape[1]):
            #         label = 'es' + str(i + 1) + str(j + 1)
            #         results[label] = effectiveStiffness[i][j]
        except KeyError:
            print('No effective stiffness matrix found.')
        try:
            indexCompliance = keywordsIndex['The Effective Compliance Matrix']
            self.eccm = utl.textToMatrix(
                linesRead[indexCompliance + 2:indexCompliance + 2 + ems])
            # for i in range(effectiveCompliance.shape[0]):
            #     for j in range(effectiveCompliance.shape[1]):
            #         label = 'ec' + str(i + 1) + str(j + 1)
            #         results[label] = effectiveCompliance[i][j]
        except KeyError:
            print('No effective compliance matrix found.')

        line = linesRead[keywordsIndex['Effective Density']]
        line = line.strip().split('=')
        self.eden = float(line[-1].strip())

        return

    def readNastranMesh(self, fn_nastran_mesh=''):
        self.fn_nastran_mesh = self.name + '.mesh'
        if fn_nastran_mesh != '':
            self.fn_nastran_mesh = fn_nastran_mesh

        lines = []
        with open(self.fn_nastran_mesh, 'r') as fin:
            entry = []
            data = ''
            for li, line in enumerate(fin):
                fmt = 'small'
                line = line.strip()
                if line == '\n' or line == '':
                    continue
                else:
                    if entry == []:
                        label = line[:8].strip()
                        if label[-1] == '*':
                            fmt = 'large'
                            label = label[:-1]
                        entry = [fmt, label]

                    if len(line) <= 72:
                        if data == '':
                            entry.append(line[8:])
                            lines.append(entry)
                            entry = []
                        else:
                            data += line[8:]
                            entry.append(data)
                            lines.append(entry)
                            data = ''
                            entry = []
                    elif len(line) > 72:
                        if data == '':
                            data = line[8:72]
                        else:
                            data += line[8:72]

        mid = 0
        for entry in lines:
            # print entry
            fmt = entry[0]
            label = entry[1]
            # data = entry[2]
            data = utl.splitNastranBulkData(entry[2], fmt)
            if label == 'GRID':  # Read GRID
                # print data
                # print data
                for i in range(2, 2 + 3):
                    if data[i] == '':
                        data[i] = '0.0'
                # print data[2:5]
                self.nodes[int(data[0])] = map(float, data[2:2 + 3])
            elif label == 'CQUAD8':  # Read CQUAD8
                self.elementids.append(int(data[0]))
                self.elements[int(data[0])] = map(int, data[2:2 + 4])
                self.elements2d[int(data[0])] = map(int, data[2:2 + 4])
                self.element_layertype[int(data[0])] = int(data[1])
            elif label == 'PSHELL':  # Read PSHELL
                self.layertypes[int(data[0])] = [int(data[1]), float(data[2])]
            elif label == 'MAT3D':  # Read MAT3D
                mid += 1
                m = Material(int(data[0]))
                m.density = float(data[-1])
                if len(data) == 11:
                    m.type = 'orthotropic'
                    m.property_elastic = map(float, data[1:10])
                self.materials[mid] = m

        for eid, nids in self.elements2d.items():
            node1 = self.nodes[nids[0]]
            node2 = self.nodes[nids[1]]
            dx2 = node2[1] - node1[1]
            dx3 = node2[2] - node1[2]
            theta1 = np.rad2deg(np.arctan2(dx3, dx2))
            if theta1 < 0.0:
                theta1 += 360.0
            self.element_theta1[eid] = theta1

        self.num_nodes = len(self.nodes)
        self.num_elements = len(self.elements)
        self.num_materials = len(self.materials)
        self.num_layertypes = len(self.layertypes)
