import os
import math
import numpy as np
import pandas as pd
# from scipy.integrate import quad, dblquad
import xml.etree.ElementTree as et
import msm.sg.sg2dcs as sg2


def calcMomentOfAreaRectangular(b, h, dx=0.0, dy=0.0):
    """ Calculate the second and polar moment of area of rectangulars

    :math:`I_x = \\frac{bh^3}{12} + Ad_x^2`

    :param b: Width of the rectangular
    :type b: float

    :param h: Height of the rectangular
    :type h: float

    :param dx: Distance between parallel x axes
    :type dx: float

    :param dy: Distance between parallel y axes
    :type dy: float

    :return: Second moments of area and polar moment of area
    :rtype: dictionary
    """
    area = b * h
    ix = b * h**3 / 12. + area * dx**2
    iy = b**3 * h / 12. + area * dy**2

    return {'ix': ix, 'iy': iy}


def calcMomentOfAreaDiskSector(r, a0, a1, center=(0.0, 0.0), origin=(0.0, 0.0)):
    """ Calculate 
    """
    a0 = np.deg2rad(a0)
    a1 = np.deg2rad(a1)
    area = (a1 - a0) / 2. * r**2
    # Centroid
    cx = 1. / 3. * r**3 * (np.sin(a1) - np.sin(a0)) / area + center[0]
    cy = 1. / 3. * r**3 * (np.cos(a0) - np.cos(a1)) / area + center[1]
    # 2nd moment of area
    ix = 1. / 8. * r**4 * ((a1 - a0) - 1. / 2. * (np.sin(2. * a1) - np.sin(2. * a0))
                           ) - area * (cy - center[1])**2 + area * (cy - origin[1])**2
    iy = 1. / 8. * r**4 * ((a1 - a0) + 1. / 2. * (np.sin(2. * a1) - np.sin(2. * a0))
                           ) - area * (cx - center[0])**2 + area * (cx - origin[0])**2

    return ix, iy


def calcMomentOfAreaAnnulusSector(ro, ri, ts, te, center=(0.0, 0.0), origin=(0.0, 0.0)):
    """ Calculate the second and polar moment of area of annulus

    :param ro: Radius of the outer circle
    :type ro: float

    :param ri: Radius of the inner circle
    :type ri: float

    :param ts: Start angle in degrees
    :type ts: float

    :param te: End angle in degrees
    :type te: float
    """
    # ts = np.deg2rad(ts)
    # te = np.deg2rad(te)
    # area = abs(te - ts) / (2. * np.pi) * np.pi * (ro**2 - ri**2)
    # ix = 1. / 16. * (ro**4 - ri**4) * (2 * (te - ts) - (np.sin(2 * te) - np.sin(2 * ts))) + area * dx**2
    # iy = 1. / 16. * (ro**4 - ri**4) * (2 * (te - ts) + (np.sin(2 * te) - np.sin(2 * ts))) + area * dy**2
    ixo, iyo = calcMomentOfAreaDiskSector(ro, ts, te, center, origin)
    ixi, iyi = calcMomentOfAreaDiskSector(ri, ts, te, center, origin)

    return ixo - ixi, iyo - iyi


def floorAbsolute(number):
    if (number - 0.) > 1e-15:
        # Positive
        number = np.floor(number)
    elif (0. - number) > 1e-15:
        # Negative
        number = np.ceil(number)
    else:
        # Very close to 0
        number = 0.

    return number


def listToString(flist, delimiter='', fmt=''):
    sfmt = '{0:' + fmt + '}'
    s = ''
    for i, n in enumerate(flist):
        if i > 0:
            s = s + delimiter + ' '
        s += sfmt.format(n)
    return s


def angleToCosine2D(angle):
    angle = math.radians(angle)
    c = [
        [math.cos(angle), -math.sin(angle)],
        [math.sin(angle),  math.cos(angle)]
    ]
    return c


def calcBasicRotation3D(angle, axis=1):
    c = np.array([
        [1.,            0.,             0.],
        [0., np.cos(angle), -np.sin(angle)],
        [0., np.sin(angle),  np.cos(angle)]
    ])
    if axis == 2:
        p = np.array([
            [0., 1., 0.],
            [0., 0., 1.],
            [1., 0., 0.]
        ])
        c = np.dot(p.T, np.dot(c, p))
    elif axis == 3:
        p = np.array([
            [0., 0., 1.],
            [1., 0., 0.],
            [0., 1., 0.]
        ])
        c = np.dot(p.T, np.dot(c, p))

    return c


def calcGeneralRotation3D(angles, order=[1, 2, 3]):
    ''' Calculate the general rotation matrix

    :param angles: Three rotation angles in radians
    :type angles: list

    :param order: The order of axis of the rotation operation
    :type order: list

    :return: The general rotation matrix
    :rtype: numpy.array
    '''
    c = np.eye(3)
    for angle, axis in zip(angles, order):
        ci = calcBasicRotation3D(angle, axis)
        c = np.dot(ci, c)
    return c


def rotateVectorByAngle2D(v2d, angle):
    c = angleToCosine2D(angle)
    vp = [
        c[0][0] * v2d[0] + c[0][1] * v2d[1],
        c[1][0] * v2d[0] + c[1][1] * v2d[1]
    ]
    return vp


def calcCab(a, b):
    '''Calculate the direction cosine matrix between frame a and b

    :math:`C_{ij} = a_i\ \cdot\ b_j`

    :param a: list of three a basis (a_1, a_2, a_3)
    :type a: list

    :param b: list of three b basis (b_1, b_2, b_3)
    :type b: list

    :return: 3a3 matrix of the direction cosine
    :rtype: numpy.array

    :Example:

    >>> a = [
    ...     [1., 0., 0.],
    ...     [0., 1., 0.],
    ...     [0., 0., 1.]
    ... ]
    >>> b = [
    ...     [],
    ...     [],
    ...     []
    ... ]
    >>> utilities.calcCab(a, b)
    '''
    a = np.array(a)
    b = np.array(b)
    Cab = np.zeros((3, 3))
    for i in range(3):
        for j in range(3):
            Cab[i][j] = np.dot(a[i], b[j])
    return Cab


def calcFramebFromSweepDihedral(sweep, dihedral):
    a1 = np.array([1., 0., 0.])
    a2 = np.array([0., 1., 0.])
    a3 = np.array([0., 0., 1.])

    dh = np.deg2rad(dihedral)
    sw = np.deg2rad(sweep)
    # 1. Dihedral (rotate around axis 2)
    Cdh = np.array([
        [np.cos(dh), 0., np.sin(dh)],
        [0., 1.,         0.],
        [-np.sin(dh), 0., np.cos(dh)]
    ])
    # 2. Sweep (rotate around axis 3)
    Csw = np.array([
        [np.cos(sw), -np.sin(sw), 0.],
        [np.sin(sw),  np.cos(sw), 0.],
        [0.,          0., 1.]
    ])

    # b1 = np.array([1., np.tan(np.deg2rad(sweep)),
    #                np.tan(np.deg2rad(dihedral))])
    # b1 = b1 / np.linalg.norm(b1)
    # b3 = np.cross(b1, a2)
    # b3 = b3 / np.linalg.norm(b3)
    # b2 = np.cross(b3, b1)
    # b2 = b2 / np.linalg.norm(b2)

    b1 = np.dot(Csw, np.dot(Cdh, a1))
    b2 = np.dot(Csw, np.dot(Cdh, a2))
    b3 = np.dot(Csw, np.dot(Cdh, a3))

    return [b1, b2, b3]


def distance(p1, p2):
    s2 = (p1[0] - p2[0])**2 + (p1[1] - p2[1])**2
    return math.sqrt(s2)


def ss(v1, v2, scale=None):
    if scale is None:
        scale = [1., ] * len(v1)
    assert len(scale) == len(v1)
    sumsqr = 0.0
    for i in range(len(v1)):
        sumsqr = sumsqr + (v1[i] * scale[i] - v2[i] * scale[i])**2
    return sumsqr


def writeFormatIntegers(file, numbers, fmt='8d', newline=True):
    ''' Write a list of integers into a file

    :param file: The file object for writing
    :type file: file

    :param numbers: The list of numbers that is going to be written
    :type numbers: list

    :param fmt: The desired format for each number
    :type fmt: string

    :param newline: If append the character ``\\n`` after writting all numbers or not
    :type newline: bool
    '''
    fmt = '{0:' + fmt + '}'
    for i in numbers:
        file.write(fmt.format(i))
    if newline:
        file.write('\n')
    return


def writeFormatFloats(file, numbers, fmt='16.7e', newline=True):
    ''' Write a list of floats into a file

    :param file: The file object for writing
    :type file: file

    :param numbers: The list of numbers that is going to be written
    :type numbers: list

    :param fmt: The desired format for each number
    :type fmt: string

    :param newline: If append the character ``\\n`` after writting all numbers or not
    :type newline: bool
    '''
    fmt = '{0:' + fmt + '}'
    for f in numbers:
        file.write(fmt.format(f))
    if newline:
        file.write('\n')
    return

# def writeDakotaParams(dictParams):
#     numVars = len(dictParams)
#     with open('params.in', 'w') as fout:
#         fout.write('{0:>32d} variables\n'.format(numVars))
#         for k, v in dictParams.items():
#             fout.write('{0:32.9E} {1:s}\n'.format(v, k))


def textToMatrix(textList):
    ''' Convert the text of a block of numbers into a matrix

    :param textList: The block of text representing a matrix
    :type textList: list

    :return: A matrix of numbers
    :rtype: numpy.array

    :Example:

    >>> lines = ['1 2 3', '4 5 6', '7 8 9']
    >>> utilities.textToMatrix(lines)
    array([[1., 2., 3.],
            [4., 5., 6.],
            [7., 8., 9.]])
    '''
    matrix = []
    for line in textList:
        line = line.strip()
        line = line.split()
        row = list(map(float, line))
        # for j in line:
        #     row.append(float(j))
        matrix.append(row)

    return np.array(matrix)


def readDakotaGAPopulation(pop, dir=r'.'):
    pop_file = dir + r'/population_' + str(pop) + r'.dat'
    pop_data = []
    with open(pop_file, 'r') as fin:
        for line in fin:
            if line.strip() == '':
                continue
            p = [pop, ] + map(float, line.strip().split())
            pop_data.append(p)
    return pop_data


def parseLayupCode(s_code):
    l_code = []
    ilsb = s_code.find('[')
    irsb = s_code.find(']')
    # print ilsb, irsb
    angles = s_code[ilsb + 1:irsb]
    # print angles
    ns = 0
    if irsb + 1 < len(s_code):
        ns = s_code[irsb + 1:].strip('s')
        if len(ns) == 0:
            ns = '1'
        ns = int(ns)
    # print ns

    # ilrbs = []
    # irrbs = []
    islash_level1 = []  # index
    is_in_brackets = False
    for i, c in enumerate(angles):
        if c == '(':
            is_in_brackets = True
            # ilrbs.append(i)
        if c == ')':
            is_in_brackets = False
            # irrbs.append(i)
        if (c == '/') and (not is_in_brackets):
            islash_level1.append(i)
    # print ilrbs
    # print irrbs
    # print islash_level1

    islash_level1.append(len(angles))
    list_angles = []
    i = 0
    for j in islash_level1:
        list_angles.append(angles[i:j])
        i = j + 1

    # print list_angles

    for i, a in enumerate(list_angles):
        temp = a.split(':')
        temp[0] = temp[0].strip('(').strip(')')
        # print temp[0]
        repeat = 1
        if len(temp) > 1:
            repeat = int(temp[-1])
        for k in range(repeat):
            if '/' in temp[0]:
                temp2 = temp[0].split('/')
                for l in temp2:
                    l_code.append(l)
            else:
                l_code.append(temp[0])

    l_code = list(map(float, l_code))
    # print l_code

    for s in range(ns):
        temp = l_code[:]
        temp.reverse()
        # print l_code
        # print temp
        l_code = l_code + temp

    # for angle in l_code:
    #     print angle

    return l_code


def updateXMLElement(parent, tag, value):
    """ Update/Create an element with given tag and value.

    :param parent: Parent element
    :type parent: xml.etree.ElementTree.Element

    :param tag: Tag of the element
    :type tag: string

    :param value: Value of the element
    :type value: string
    """
    _se = parent.find(tag)
    if _se is None:
        _se = et.SubElement(parent, tag)
    _se.text = value


def splitNastranBulkData(data_string, fmt='small'):
    data_list = []
    nchar = 8
    if fmt == 'large':
        nchar = 16
    start = 0
    # print len(data_string)
    while (start + nchar) <= len(data_string):
        # print start + nchar
        # print data_string[start:start + nchar]
        data_list.append(data_string[start:start + nchar].strip())
        start += nchar
    data_list.append(data_string[start:])

    return data_list


def sortGEBTStepResultStatic(gebt_step_result):
    """ Rearrange GEBT results of points and members of a simply beam.

    Sort by reference line coordinate.
    """
    result_sorted = np.array([gebt_step_result['points'][0],])
    for i in range(len(gebt_step_result['members'])):
        result_sorted = np.vstack((
            result_sorted,
            gebt_step_result['members'][i],
            gebt_step_result['points'][i+1]
        ))
    return result_sorted


def calcBeamRefLineResultFromDNS(
    fn_abaqus_rpt, col_labels,
    skiptop=0, skipbottom=0, tolerance=1e-6
):
    """ Read Abaqus 3D DNS results from a .rpt file,
    and calculate the corresponding reference line results.

    This is mainly for beam-like hollow structures, where the reference 
    line is not modeled by any elements.

    Assume the first column is the node label (integer) and the rest 
    are quantities of interests (floats).

    The (un)deformed reference line is calculated as the average of the 
    (un)deformed coordinates (a2/a3) of nodes at each spanwise loacation (a1).
    """

    df_raw = pd.read_csv(
        fn_abaqus_rpt,
        delim_whitespace=True, header=None, names=col_labels,
        skiprows=skiptop, skipfooter=skipbottom, engine='python'
    )

    # Calculate undeformed and deformed reference line coordinates
    df_raw['a2d'] = df_raw['a2'] + df_raw['u2']
    df_raw['a3d'] = df_raw['a3'] + df_raw['u3']
    df_grouped = df_raw.groupby('a1')
    df_ref = df_grouped[
        'a2', 'a3', 'u1', 'u2', 'u3', 'a2d', 'a3d'
    ].mean()
    df_ref.columns = ['o2', 'o3', 'ou1', 'ou2', 'ou3', 'o2d', 'o3d']

    # print(df_averaged.loc[0.0, :])

    temp = df_raw.set_index('a1')
    temp = pd.concat([temp, df_ref], axis=1)
    # Remove small values
    temp['a2t'] = temp['a2'].apply(lambda x: np.nan if np.abs(x) < tolerance else x)
    temp['a3t'] = temp['a3'].apply(lambda x: np.nan if np.abs(x) < tolerance else x)
    # Relative displacement with respect to the reference line
    temp['ru2'] = temp['u2'] - temp['ou2']
    temp['ru3'] = temp['u3'] - temp['ou3']
    # Calculate rotation angles
    temp['r23'] = -temp['ru2'] / temp['a3t']
    temp['r32'] = temp['ru3'] / temp['a2t']

    temp_grouped = temp.groupby('a1')
    temp2 = temp_grouped['r23', 'r32'].mean()
    temp2['r1'] = (temp2['r23'] + temp2['r32']) / 2.0

    df_final = pd.concat([df_ref, temp2.loc[:, 'r1']], axis=1).reset_index()

    return df_final


def mergeNodes(fn_in_old, fn_in_new=None, format='vabs', tol=1e-6):
    """
    Merge overlaped nodes having distance smaller than the tolerance.

    :param fn_in_old: Input file that needs to be merged.
    :type fn_in_old: string

    :param fn_in_new: The new input file.
    :type fn_in_new: string

    :param format: VABS ('vabs') or SwiftComp ('swiftcomp').
    :type format: string

    :param tol: Tolerance for checking the distance.
    :type tol: float
    """
    if fn_in_new is None:
        fn_in_new = fn_in_old

    cs = sg2.SGCrossSection('temp')
    print('> Reading old input file...')
    cs.readVABSIn(fn_vabs_in=fn_in_old)

    nodes = []
    for k, v in cs.nodes.items():
        nodes.append((k, int(v[1] / tol), int(v[2] / tol)))
    dtype = [('id', int), ('y2', int), ('y3', int)]
    nodes = np.array(nodes, dtype=dtype)

    print('> Sorting nodes...')
    nodes = np.sort(nodes, order=['y2', 'y3', 'id'])

    print('> Checking distance and assigning new id...')
    old2new_map = {}
    new_nodes = {}
    newid = 0
    count = 0
    nnodes_removed = 0
    for i, n in enumerate(nodes):
        count += 1
        if count % 10000 == 0:
            print('  > progress: {0:d} / {1:d}'.format(count, cs.num_nodes))
        if i == 0:
            newid += 1
            old2new_map[n[0]] = newid
            new_nodes[newid] = cs.nodes[n[0]]
        else:
            prev = nodes[i - 1]
            if (n[1] == prev[1]) and (n[2] == prev[2]):
                nnodes_removed += 1
                old2new_map[n[0]] = old2new_map[prev[0]]
            else:
                newid += 1
                old2new_map[n[0]] = newid
                new_nodes[newid] = cs.nodes[n[0]]
    print('  > done.')
    print('  Number of removed nodes: {0:d}'.format(nnodes_removed))
    cs.num_nodes = newid

    print('> Updating element connectivities...')
    new_elements = {}
    count = 0
    for eid, nodes in cs.elements.items():
        count += 1
        if count % 10000 == 0:
            print('  > progress: {0:d} / {1:d}'.format(count, cs.num_elements))
        new_elements[eid] = []
        for nid in nodes:
            new_elements[eid].append(old2new_map[nid])
    print('  > done.')

    cs.nodes = new_nodes
    cs.elements = new_elements
    cs.elements2d = new_elements

    print('> Writing new input file...')
    cs.writeVABSIn(fn_vabs_in=fn_in_new)

    print('> FINISH')
